package Language::INTERCAL::Registers;

# Bytecode sequences to encode registers and related functions

# This file is part of CLC-INTERCAL

# Copyright (c) 2006-2008, 2023 Claudio Calvelli, all rights reserved.

# CLC-INTERCAL is copyrighted software. However, permission to use, modify,
# and distribute it is granted provided that the conditions set out in the
# licence agreement are met. See files README and COPYING in the distribution.


use strict;
use Carp;
use vars qw($VERSION $PERVERSION $DATAVERSION);
($VERSION) = ($PERVERSION = "CLC-INTERCAL/Base INTERCAL/Registers.pm 1.-94.-2.4") =~ /\s(\S+)$/;

use Language::INTERCAL::RegTypes '1.-94.-2.2',
    qw(REG_spot REG_twospot REG_tail REG_hybrid REG_dos REG_whp REG_shf REG_cho reg_nametype reg_typename);
use Language::INTERCAL::Exporter '1.-94.-2', qw(import compare_version);
use Language::INTERCAL::GenericIO '1.-94.-2',
	qw($stdwrite $stdread $stdsplat $devnull);
use Language::INTERCAL::Splats '1.-94.-2.1', qw(faint SP_SPECIAL);
use Language::INTERCAL::ByteCode '1.-94.-2.2', qw(BC);
use Language::INTERCAL::DoubleOhSeven '1.-94.-2.2', qw(make_doubleohseven);
use Language::INTERCAL::SharkFin '1.-94.-2.2', qw(make_sharkfin);

$DATAVERSION = '1.-94.-2.5';
compare_version($VERSION, $DATAVERSION) < 0 and $VERSION = $DATAVERSION;

use vars qw(@EXPORT_OK);
# for simplicity we re-export REG_* from RegTypes
@EXPORT_OK = qw(
    REG_spot REG_twospot REG_tail REG_hybrid REG_dos REG_whp REG_shf REG_cho
    reg_nametype reg_typename
    add_register reg_code reg_code2 reg_create reg_decode reg_list
    reg_name reg_translate all_registers
    DOS_AR DOS_AW DOS_BA DOS_CF DOS_CR DOS_CW DOS_DM DOS_ES DOS_FS DOS_GU
    DOS_IO DOS_IS DOS_JS DOS_NB DOS_OS DOS_PS DOS_RM DOS_RT DOS_SM DOS_SP
    DOS_SS DOS_TM DOS_WT
    SHF_AV SHF_EV
    WHP_OR WHP_ORFH WHP_OSFH WHP_OWFH WHP_SNFH WHP_TRFH
);

# these are duplicated from ByteCode.pm so we don't need a mutual dependency
# they are autongenerated from ByteCode.Data anywa
sub BC_CHO () { 71; }
sub BC_DOS () { 69; }
sub BC_HYB () { 67; }
sub BC_SHF () { 70; }
sub BC_SPO () { 64; }
sub BC_TAI () { 66; }
sub BC_TSP () { 65; }
sub BC_TYP () { 79; }
sub BC_WHP () { 68; }

my @reg_list = qw(
    AR AV AW BA CF CR CW DM ES EV FS GU IO IS JS NB OR ORFH OS OSFH OWFH PS
    RM RT SM SNFH SP SS TM TRFH WT
);

my %reg_list = (
    AR => ['spot', 0, BC_DOS, '%', 10, 'Array read value', "Contains the last byte READ OUT when %IO is C."],
    AW => ['spot', 0, BC_DOS, '%', 11, 'Array write value', "Contains the last byte WRITten IN when %IO is C."],
    BA => ['base', 2, BC_DOS, '%', 4, 'BAse', "Holds the base used for all arithmetic. Assigning a value less than 2 or greater than 7 causes an error."],
    CF => ['comefrom', 0, BC_DOS, '%', 5, 'Come From style', "This register can only hold values from #0 to #3. The lowest bit (in base 2) determines what happens when multiple COME FROM or NEXT FROM all point at the same label: if zero, you get a splat, if one you get a multithreaded program. The other bit determines whether COME FROM gerund (and NEXT FROM gerund) will work: zero disables these statements, one enables them. Thus all compilers set %CF to #0, except I<thick.io> which sets it to #1 and I<come-from-gerund.io> which sets it to #2 (if both these objects are preloaded, they work together to set %CF to #3)."],
    CR => ['charset', 0, BC_DOS, '%', 6, 'Charset for Reads', "The character set used by alphanumeric READ OUT when %IO is CLC. Assigning a number to this register causes the corresponding character set to be selected; assigning a I<MUL> causes a symbolic lookup to determine the character set number. See L<Language::INTERCAL::Charset>."],
    CW => ['charset', 0, BC_DOS, '%', 7, 'Charset for Writes', "The character set used by alphanumeric WRITE IN when %IO is CLC. Assigning a number to this register causes the corresponding character set to be selected; assigning a I<MUL> causes a symbolic lookup to determine the character set number. See L<Language::INTERCAL::Charset>."],
    DM => ['zeroone', 0, BC_DOS, '%', 18, 'unary Division mode', "This register can only have value #0 or #1, and determines the style of unary division employed. The default value #0 corresponds to the \"arithmetic\" style of division, while value #1 corresponds to the \"bitwise\" style. See the I<UDV> opcode."],
    ES => ['symbol', 'CALC_EXPR', BC_DOS, '%', 16, '"expr" symbol', "Used by the Intercal calculator (intercalc) to determine how to parse lines when operating in \"expr\" mode."],
    FS => ['symbol', 'CALC_FULL', BC_DOS, '%', 15, '"full" symbol', "Used by the Intercal calculator (intercalc) to determine how to parse lines when operating in \"full\" mode."],
    GU => ['zeroone', 0, BC_DOS, '%', 23, 'abstain Give Up', "This register can only have value #0 or #1, and determines whether an ABSTAIN FROM (label) or the corresponding REINSTATE acts on a GIVE UP at that label. CLC-INTERCAL traditionally does not allow this, but this register can be set to have behaviour compatible with C-INTERCAL, which does allow it."],
    IO => ['iotype', 1, BC_DOS, '%', 3, 'I/O type', "Determines how non-numeric WRITE IN and READ OUT work. Assigning a number to this register causes the corresponding style to be selected; assigning a I<MUL> causes a symbolic lookup to determine the style number. See L<Language::INTERCAL::ArrayIO>."],
    IS => ['symbol', 0, BC_DOS, '%', 17, 'Intersection symbol', "Determines what separates statements in the program; this is not currently used in any compiler and can be left at the default, zero. If set to any other value, the corresponding grammar symbol is used to compile the bit of source between consecutive statements; if this generates code, it will be executed with the preceding statement. Changing the value of this register at runtime can be a great obfuscation tool. See also I<%PS>."],
    JS => ['symbol', 'END_JUNK', BC_DOS, '%', 12, 'Junk Symbol', "When parsing a comment, the compiler needs to be told how to recognise the start of next statement: for example, I<1972.io> and I<ick.io> set this to a grammar symbol meaning \"optional (number) followed by DO or PLEASE\"; I<sick.io> does something similar, but the complication caused by computed labels (if enabled) makes it alightly more difficult to describe what this symbol does.\n\nChanging the value of this register at runtime can be a great obfuscation tool."],
    NB => ['zeroone', 0, BC_DOS, '%', 24, 'NonBlocking', "This register can only have value #0 or #1, and determines whether a blocking operation stops all threads or only suspends the current thread; before 1.-94.-2.5 this register did not exist, and blocking operations always stopped all threads."],
    OS => ['spot', 0, BC_DOS, '%', 8, 'Operating System', "Hidden in the darkest corner of the operating system lurks a \"DO NEXT FROM %OS\". As long as %OS has the default value of zero, you are safe from this.\n\nIf %OS is assigned some other value, it behaves like a normal (?) NEXT FROM, with one added twist to do with parameter passing. Every time your program assigns a value to a register, %OS will be made to belong to the register being assigned to, and any previous belonging removed. This allows the system call code to refer to I<\$%OS> to try to guess what you want. The system call will use up to five arguments, provided by registers I<.-\$%OS>, I<:-\$%OS>, I<,-\$%OS>, I<;-\$%OS> and I<\@%\$OS>, in other words the spot, two spot, tail, hybrid and whirlpool register with the same number as whatever you last assigned to. This is called \"call by vague resemblance to the last assignment\" and, to our knowledge, no other language has ever used this style of parameter passing.\n\nTo use, you do something like \"(666) DO .5 <- #1\" which would execute syscall #1, assuming %OS has the value #666. This particular example would store the version number of CLC-INTERCAL in ,5."],
    PS => ['symbol', 'PROGRAM', BC_DOS, '%', 14, 'Program symbol', "Determines where the compiler starts when parsing a program. This should be a grammar symbol corresponding to a single statement, the symbol is automaticaly used repeatedly to parse the whole program. Changing the value of this register at runtime can be a great obfuscation tool. See also I<%IS>."],
    RM => ['zeroone', 0, BC_DOS, '%', 21, 'Reinstate Mode', "This register can only have value #0 or #1, and determines whether a REINSTATE of an IGNOREd register behaves in the traditional way (#0) or in the way documented by the CLC-INTERCAL documentation (#1)."],
    RT => ['roman', 1, BC_DOS, '%', 2, 'Read Type', "Determines how numeric READ OUT produces its output. Assigning a number to this register causes the corresponding style to be selected; assigning a I<MUL> causes a symbolic lookup to determine the style number. See L<Language::INTERCAL::ReadNumbers>."],
    SM => ['zeroone', 1, BC_DOS, '%', 22, 'Select Mode', "This register can only have value #0 or #1, and determines the size of the result of a select binary operator: #1 means that the size of the result is data-dependent: if it fits in one spot, it's one spot, otherwise it's two spot: this is the normal CLC-INTERCAL behaviour; #0 means that the result has the size of the second operant: this is the C-INTERCAL behaviour."],
    SP => ['splat', 1000, BC_DOS, '%', 19, 'SPlat', "This register contains the code of the last splat, just like the '*' expression. Assigning to it, however, does not cause a splat, but will trigger any events depending on splats. This register is intended for internal use by the compiler; programs should use I<SPL> instead."],
    SS => ['symbol', 'SPACE', BC_DOS, '%', 13, 'Space symbol', "The compiler will automatically ignore anything matched by this symbol. If the Whitespace extension is installed, anything matched by this symbol is passed to the Whitespace compiler. See the documentation which comes with the Whitespace extension. Changing the value of this register at runtime can be a great obfuscation tool."],
    TM => ['zeroone', 0, BC_DOS, '%', 9, 'Trace Mode', "If %TM is zero, the program is not traced. If it is #1 the program will send bytecode trace information to \@TRFH. Assigning any other value to %TM is an error."],
    WT => ['zeroone', 0, BC_DOS, '%', 1, 'Write Type', "Determines how numeric WRITE IN behaves. The default value of #0 corresponds to the standard, traditional form; the value #1 enables wimp mode for input. Any other value is invalid."],
    AV => ['vector', [], BC_SHF, '^', 1, 'Argument Vector', "The runtime initialises it with any command-line arguments provided to the program."],
    EV => ['vector', [], BC_SHF, '^', 2, 'Environment Vector', "The runtime initialises it with environment variables available when the program starts."],
    OR => ['whirlpool', undef, BC_WHP, '@', 0, 'Overload Register', "This register cannot be used directly. During overload, it belongs to the register being overloaded, so it's possible to access that register using I<BLM> I<1> I<WHP> I<0>, or the corresponding INTERCAL source (I<\$\@0>). Any other use will probably splat."],
    ORFH => ['whirlpool', $stdread, BC_WHP, '@', 2, 'Object\'s read filehandle', "Used by default for all READ OUT statements. Defaults to the standard read filehandle, which your operating system may call \"standard output\""],
    OSFH => ['whirlpool', $stdsplat, BC_WHP, '@', 3, 'Object\'s splat filehandle', "Used by default for the text produced by a splat. Defaults to the standard splat filehandle, which your operating system may call \"standard error\"."],
    OWFH => ['whirlpool', $stdwrite, BC_WHP, '@', 1, 'Object\'s write filehandle', "Used by default for all WRITE IN statements. Defaults to the standard write filehandle, which your operating system may call \"standard input\"."],
    SNFH => ['whirlpool', $devnull, BC_WHP, '@', 7, 'Null filehandle', "If you READ OUT to \@7 nothing happens. If you WRITE IN from \@7 you get an empty write (for numeric writes, a value of #0, for non-numeric writes an empty array)."],
    TRFH => ['whirlpool', $stdsplat, BC_WHP, '@', 9, 'Trace filehandle', "Trace output, if enabled by I<%TM>, is sent here. Defaults to the standard splat filehandle, which your operating system may call \"standard error\""],
);

my %reg_names = (
    '%1' => 'WT',
    '%2' => 'RT',
    '%3' => 'IO',
    '%4' => 'BA',
    '%5' => 'CF',
    '%6' => 'CR',
    '%7' => 'CW',
    '%8' => 'OS',
    '%9' => 'TM',
    '%10' => 'AR',
    '%11' => 'AW',
    '%12' => 'JS',
    '%13' => 'SS',
    '%14' => 'PS',
    '%15' => 'FS',
    '%16' => 'ES',
    '%17' => 'IS',
    '%18' => 'DM',
    '%19' => 'SP',
    '%21' => 'RM',
    '%22' => 'SM',
    '%23' => 'GU',
    '%24' => 'NB',
    '^1' => 'AV',
    '^2' => 'EV',
    '@0' => 'OR',
    '@1' => 'OWFH',
    '@2' => 'ORFH',
    '@3' => 'OSFH',
    '@7' => 'SNFH',
    '@9' => 'TRFH',
);

sub DOS_AR () { 10 }
sub DOS_AW () { 11 }
sub DOS_BA () { 4 }
sub DOS_CF () { 5 }
sub DOS_CR () { 6 }
sub DOS_CW () { 7 }
sub DOS_DM () { 18 }
sub DOS_ES () { 16 }
sub DOS_FS () { 15 }
sub DOS_GU () { 23 }
sub DOS_IO () { 3 }
sub DOS_IS () { 17 }
sub DOS_JS () { 12 }
sub DOS_NB () { 24 }
sub DOS_OS () { 8 }
sub DOS_PS () { 14 }
sub DOS_RM () { 21 }
sub DOS_RT () { 2 }
sub DOS_SM () { 22 }
sub DOS_SP () { 19 }
sub DOS_SS () { 13 }
sub DOS_TM () { 9 }
sub DOS_WT () { 1 }
sub SHF_AV () { 1 }
sub SHF_EV () { 2 }
sub WHP_OR () { 0 }
sub WHP_ORFH () { 2 }
sub WHP_OSFH () { 3 }
sub WHP_OWFH () { 1 }
sub WHP_SNFH () { 7 }
sub WHP_TRFH () { 9 }

# line 69

sub add_register {
    @_ == 8 or croak
	"Usage: add_register(SOURCE, NAME, TYPE, CODE, NUMBER, DEFAULT, DESCR, DOC)";
    my ($source, $name, $type, $code, $number, $default, $descr, $doc) = @_;
    my $op;
    if ($type eq '%') {
	$op = BC_DOS;
    } elsif ($type eq '^') {
	$op = BC_SHF;
    } elsif ($type eq '@') {
	$op = BC_WHP;
    } else {
	croak "Invalid TYPE $type";
    }
    $name = uc($name);
    exists $reg_list{$name}
	and croak "Duplicate register name: $name";
    $number += 0;
    exists $reg_names{$type . $number}
	and croak "Duplicate register number: $type$number";
    $reg_list{$name} = [$code, $default, $op, $type, $number, $descr, $doc, $source];
    $reg_names{$type . $number} = $name;
    push @reg_list, $name;
}

sub reg_decode ($$;$) {
    my ($type, $number, $whp) = @_;
    my $typename = reg_typename($type) || "?$type/";
    my $rn = "$typename$number";
    $type == REG_dos || $type == REG_shf || ($whp && $type == REG_whp)
	or return $rn;
    exists $reg_names{$rn} or return $rn;
    $rn = $reg_names{$rn};
    $reg_list{$rn}[3] . $rn;
}

sub reg_create ($$$) {
    my ($type, $number, $object) = @_;
    if ($type == REG_dos) {
	if (exists $reg_names{"%$number"}) {
	    my $rn = $reg_names{"%$number"};
	    exists $reg_list{$rn}
		and return make_doubleohseven($reg_list{$rn}[0],
					      $object,
					      $reg_list{$rn}[1],
					      REG_spot);
	}
	faint(SP_SPECIAL, reg_decode($type, $number));
    }
    if ($type == REG_shf) {
	if (exists $reg_names{"^$number"}) {
	    my $rn = $reg_names{"^$number"};
	    exists $reg_list{$rn}
		and return make_sharkfin($reg_list{$rn}[0],
					 $object,
					 $reg_list{$rn}[1],
					 REG_tail);
	}
	faint(SP_SPECIAL, reg_decode($type, $number));
    }
    if ($type == REG_whp) {
	if (exists $reg_names{"\@$number"}) {
	    my $rn = $reg_names{"\@$number"};
	    exists $reg_list{$rn}
		and return ( { filehandle => $reg_list{$rn}[1] } );
	}
	return ( {} );
    }
    $type == REG_spot || $type == REG_twospot
	and return (0);
    $type == REG_tail || $type == REG_hybrid
	and return [];
    $type == REG_cho && ($number == 1 || $number == 2)
	and return ();
    faint(SP_SPECIAL, reg_decode($type, $number));
}

sub reg_list () {
    @reg_list;
}

sub reg_code ($) {
    my ($name) = @_;
    my ($type, $number) = reg_translate($name);
    reg_code2($type, $number);
}

sub reg_code2 ($$) {
    my ($type, $number) = @_;
    $type == REG_spot and return (BC_SPO, BC($number));
    $type == REG_twospot and return (BC_TSP, BC($number));
    $type == REG_tail and return (BC_TAI, BC($number));
    $type == REG_hybrid and return (BC_HYB, BC($number));
    $type == REG_whp and return (BC_WHP, BC($number));
    $type == REG_dos and return (BC_DOS, BC($number));
    $type == REG_shf and return (BC_SHF, BC($number));
    $type == REG_dos and return (BC_CHO, BC($number));
    undef;
}

sub reg_name ($) {
    my ($rn) = @_;
    exists $reg_list{$rn}
	and return $reg_list{$rn}[3] . $reg_list{$rn}[4];
    if (exists $reg_names{$rn}) {
	$rn = $reg_names{$rn};
	return $reg_list{$rn}[3] . $reg_list{$rn}[4];
    }
    $rn =~ /^([%^\@])(.*)$/ && exists $reg_list{$2} && $reg_list{$2}[3] eq $1
	and return $reg_list{$2}[3] . $reg_list{$2}[4];
    $rn =~ s/^([\.:,;\@^%])0*([^\D0]\d*)$/$1$2/ && $2 <= 0xffff and return $rn;
    $rn =~ /^\@0+$/ and return '@0';
    undef;
}

sub reg_translate ($) {
    my ($rn) = @_;
    $rn = $reg_names{$rn} if exists $reg_names{$rn};
    exists $reg_list{substr($rn, 1)} && $reg_list{substr($rn, 1)}[3] eq substr($rn, 0, 1)
	and $rn = substr($rn, 1);
    if (exists $reg_list{$rn}) {
	my $rt = $reg_list{$rn}[3];
	my $rv = $reg_list{$rn}[4];
	$rt eq '%' and return (REG_dos, $rv);
	$rt eq '^' and return (REG_shf, $rv);
	$rt eq '@' and return (REG_whp, $rv);
    }
    my $rt = substr($rn, 0, 1);
    my $rv = 0 + substr($rn, 1);
    $rv > 0 && $rv <= 0xffff or faint(SP_SPECIAL, $rn);
    $rt eq '.' and return (REG_spot, $rv);
    $rt eq ':' and return (REG_twospot, $rv);
    $rt eq ',' and return (REG_tail, $rv);
    $rt eq ';' and return (REG_hybrid, $rv);
    $rt eq '@' and return (REG_whp, $rv);
    $rt eq '_' and return (REG_cho, $rv);
    faint(SP_SPECIAL, $rn);
}

sub all_registers {
    map { [$_, @{$reg_list{$_}}] } keys %reg_list;
}

1;
